/*
 * OPALE is a scientific library under LGPL. Its main goal is to
 * develop mathematical tools for any scientist.
 *
 * Copyright (C) 2002 Opale Group
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * You can visit the web site http://opale.tuxfamily.org to obtain more
 * informations about this program and/or to contact the authors by mail
 * developers@opale.tuxfamily.org.
 */

package opale.mathtools;
import opale.tools.*;

import java.io.*;
import java.util.*;

/**
* This class extends IntegrationRule and implements the Trapezoidal rule to integrate a function.
* @since Opale-Mathtools 0.18
*/
public final class TrapezoidalRule extends IntegrationRule
{
/**
* This value represents the precision of the solution.
*/
public double PRECISION = 1.e-6;

/**
* This value fixes the maximum number of iterations in the algorithm.
*/ 
public int MAXITERATION = 20;
/**
* This variable contains the number of iterations after convergence.
*/ 
public int NITER = 0;

/**
* Evaluate the sum of a specified function between limits  a and b by the trapezoidal rule. The used algorithm can be found in the numerical recipes book.
* @param IyFx f, a function.
* @param double a, b, limits of integration interval.
* @param int ni, this value indicates the number of interval to be used to estimate the sum.
* @return double, the result.
*/ 
public double integrate(IyFx f, double a, double b, int ni) 
{
if (b-a<=0.) throw new IllegalArgumentException("b-a must be a positive number !!");
if (ni<=0) throw new IllegalArgumentException("ni must be a positive integer !!");

double sum = 0.;
double olds = -1.e30;
ni = (int) Math.round((Math.log(ni+1)/Math.log(2.)));
NITER=0;

if (ni>MAXITERATION) throw new IllegalArgumentException("Too many intervals !!");

while (NITER<ni)
	{
	sum = trapz(sum,f,a,b,NITER);
	if (NITER>5)
		if ( (Math.abs(sum-olds)<PRECISION*Math.abs(olds)) || ( (sum==0.) && (olds ==0.) ) ) return sum;
	olds = sum;
	NITER++;
	}

return sum;
}

/**
* This routine performs the nth stage of refinement of extended Trapzeoidal rule, according to numerical recipes book.
*/
private double trapz(double s,IyFx f, double a, double b, int n)
{

if (n==1) return 0.5*(b-a)*(f.val(a)+f.val(b));
else
	{
	int nit = 1;
	double dnit;
	double sum=0.;
	double dx = b-a;
	double x;
	
	for (int i=1;i<n-1;i++) nit<<=1; //number of interval 
	dnit = (double) nit;
	dx /= dnit;
	x = a+0.5*dx;
	for (int i=0;i<nit;i++)
		{
		sum+=f.val(x);
		x+=dx;
		}
	s = 0.5*(s+dx*sum);
	return s;
	}
//return trapz(s,f, a,  b, n); // not reached !!
	
}

}
